/**
 * Copyright (c) 2007-2012 EBM WebSourcing, 2012-2017 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.databinding.jaxb.serialization;


import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.lang.reflect.ParameterizedType;
import java.lang.reflect.Type;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.List;
import java.util.Map;

import javax.xml.bind.JAXBElement;
import javax.xml.bind.Marshaller;
import javax.xml.bind.annotation.XmlElement;
import javax.xml.bind.annotation.XmlElementRef;
import javax.xml.bind.annotation.XmlRootElement;
import javax.xml.bind.annotation.XmlType;
import javax.xml.datatype.DatatypeConfigurationException;
import javax.xml.datatype.DatatypeFactory;
import javax.xml.datatype.XMLGregorianCalendar;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.ow2.petals.databinding.jaxb.DataBindingException;
import org.ow2.petals.databinding.jaxb.Registry;
import org.ow2.petals.databinding.jaxb.service.Message;
import org.ow2.petals.databinding.jaxb.service.Operation;
import org.ow2.petals.databinding.jaxb.tools.DataBindingConstants;
import org.ow2.petals.databinding.jaxb.tools.Util;
import org.w3c.dom.Document;

/**
 * This class permits to marshal the request/response message of the service operations.
 * @author Mohammed EL JAI - EBM WebSourcing
 */
public class ServiceMarshaller {

    /**  The class loader of the generated JAXB classes **/
    private ClassLoader jaxbClassLoader;

    /** **/
    private Map<String, String> classPackageMapping;

    /**
     * Marshall the method result of the service operation given in parameter
     * @param serviceOperation : The service operation 
     * @param methodResult : The method result 
     * @return a XML message response of the operation given in parameter 
     * @throws DataBindingException
     */
    @SuppressWarnings("unchecked")
    public Document marshallMethodResult(Operation serviceOperation, Object[] methodResult) 
    throws DataBindingException  {

        DocumentBuilderFactory pFactory = DocumentBuilderFactory.newInstance();
        Document result =null;
        //initializing result
        try {
            result = pFactory.newDocumentBuilder().newDocument();
        } catch (ParserConfigurationException e1) {
            throw new DataBindingException(e1);
        }   


        Message outputMessage = serviceOperation.getOutputMessage();
        String messageNameSpace = outputMessage.getMessageTypeName().getNamespaceURI();
        Registry registry = serviceOperation.getServiceOfOperation().getRegistry();
        Marshaller marshaller = registry.getMarshaller(messageNameSpace);
        Object messageJaxbObject = null;

        // Get output message info
        Map<String,Class<?>> messageParts = outputMessage.getMessageParts();
        String methodName = "create" + Util.upperFirstCase(outputMessage.getMessageTypeName().getLocalPart());

        // Get objects needed to marshalling message operation
        Class<?> jaxbFactoryClass = registry.getJaxbFactoryInstance(messageNameSpace).getClass();
        Object jaxbFactoryInstance = registry.getJaxbFactoryInstance(messageNameSpace);


        try {
            Method createJaxbObject = jaxbFactoryClass.getDeclaredMethod(methodName, new Class[]{});
            messageJaxbObject = createJaxbObject.invoke(jaxbFactoryInstance, new Object[]{});
        } catch (Exception e1) {
            throw new DataBindingException("Error occured when creating the message object", e1);
        } 

        if (serviceOperation.getReturnType() != null && serviceOperation.getReturnType().getName().
                equalsIgnoreCase(serviceOperation.getOutputMessage().getMessageJaxbClass().getName())) {
            Object tempMessageJaxbObject = createJaxblBean(methodResult[0]);
            if (tempMessageJaxbObject != null)
                messageJaxbObject = tempMessageJaxbObject;

        }
        else

        {

            try {


                Class<?> messageJaxbClass = outputMessage.getMessageJaxbClass();
                Class<?> returnType = serviceOperation.getReturnType();



                // Browse the method result
                int messageParamIndex = 0;
                for (String messgePartName : messageParts.keySet()) {

                    String attributeName = null;
                    if (messageJaxbClass.getDeclaredFields()[messageParamIndex].getAnnotation(XmlElementRef.class) != null
                            && messageJaxbClass.getDeclaredFields()[messageParamIndex].getAnnotation(XmlElementRef.class).name() != null)
                        attributeName = messageJaxbClass.getDeclaredFields()[messageParamIndex].getAnnotation(XmlElementRef.class).name();
                    else 
                        attributeName = messageJaxbClass.getDeclaredFields()[messageParamIndex].getName();
                    String createMethodName = "create" + messageJaxbClass.getSimpleName() 
                    + Util.upperFirstCase(attributeName);


                    Class<?> messagePartClass = messageParts.get(messgePartName);
                    Method setterMethod = null;
                    // If the field is a collection type, there is no a setter method
                    if (!java.util.Collection.class.isAssignableFrom(messagePartClass)) {
                        String setterMethodName = "set" + Util.upperFirstCase(messgePartName);
                        setterMethod = messageJaxbClass.getDeclaredMethod(setterMethodName, 
                                new Class[]{messagePartClass});
                    }

                    if (methodResult[messageParamIndex] == null) {
                        setterMethod.invoke(messageJaxbObject, (Object) null);
                    }

                    // The method result is a simple type or primitive type
                    else if (returnType.isPrimitive() || Util.isSimpleType(returnType)) {

                        Method createMethod = null;
                        if (messageJaxbClass.getDeclaredFields()[messageParamIndex].getType().equals(JAXBElement.class)) 
                            createMethod = jaxbFactoryClass.getDeclaredMethod(createMethodName, 
                                    returnType);

                        if (createMethod != null) 
                            setterMethod.invoke(messageJaxbObject, 
                                    createMethod.invoke(jaxbFactoryInstance, methodResult[messageParamIndex]));

                        else
                            setterMethod.invoke(messageJaxbObject, new Object[]{methodResult[messageParamIndex]});

                    }

                    // The method result is a Calendar type
                    else if (returnType.getName().equals(DataBindingConstants.CALENDAR_CLASS_NAME)) {

                        javax.xml.datatype.XMLGregorianCalendar jaxbDate = null;
                        java.util.Calendar originalBeanDate = (java.util.Calendar) methodResult[messageParamIndex];
                        if (originalBeanDate != null) {
                            jaxbDate = createXMLGregorianCalendarFromDate(originalBeanDate.getTime());
                        }

                        Method createMethod = null;
                        if (messageJaxbClass.getDeclaredFields()[messageParamIndex].getType().equals(JAXBElement.class)) 
                            createMethod = jaxbFactoryClass.getDeclaredMethod(createMethodName, 
                                    javax.xml.datatype.XMLGregorianCalendar.class);

                        if (createMethod != null) 
                            setterMethod.invoke(messageJaxbObject,
                                    createMethod.invoke(jaxbFactoryInstance,jaxbDate));
                        else
                            setterMethod.invoke(messageJaxbObject, new Object[]{jaxbDate});

                    }

                    // The method result is a Date type
                    else if (returnType.getName().equals(DataBindingConstants.DATE_CLASS_NAME) ||
                            returnType.getName().equals(DataBindingConstants.SQL_DATE_CLASS_NAME)) {

                        javax.xml.datatype.XMLGregorianCalendar jaxbDate = null;
                        java.util.Date originalBeanDate = (java.util.Date) methodResult[messageParamIndex];
                        if (originalBeanDate != null) {
                            jaxbDate = createXMLGregorianCalendarFromDate(originalBeanDate);
                        }
                        Method createMethod = null;
                        if (messageJaxbClass.getDeclaredFields()[messageParamIndex].getType().equals(JAXBElement.class)) 
                            createMethod = jaxbFactoryClass.getDeclaredMethod(createMethodName, 
                                    javax.xml.datatype.XMLGregorianCalendar.class);

                        if (createMethod != null) 
                            setterMethod.invoke(messageJaxbObject,
                                    createMethod.invoke(jaxbFactoryInstance,jaxbDate));
                        else
                            setterMethod.invoke(messageJaxbObject, new Object[]{jaxbDate});

                    }

                    // The method result is a collection
                    else if (java.util.Collection.class.isAssignableFrom(returnType) || returnType.isArray()) {

                        Object jaxbList = null;
                        Field messageField = Util.getJaxbField(messgePartName, messageJaxbClass.getDeclaredFields());


                        if (java.util.Collection.class.isAssignableFrom(messageField.getType())) {
                            if (setterMethod != null)
                                jaxbList = new ArrayList<Object>();
                            else {
                                String getterMethodName = "get" + Util.upperFirstCase(messgePartName);
                                Method getterMethod = messageJaxbClass.getDeclaredMethod(getterMethodName);
                                jaxbList = getterMethod.invoke(messageJaxbObject);
                            }
                            if (methodResult[messageParamIndex].getClass().isArray()) {
                                try {
                                    createJaxbList(java.util.Arrays.asList((Object[]) methodResult[messageParamIndex]),jaxbList);
                                }
                                catch (java.lang.ClassCastException e) {
                                    // The array contains a primitive type
                                    List<?> listOfWrappedClass = Util.parimitiveArrayToList(methodResult[messageParamIndex]);
                                    createJaxbList(listOfWrappedClass,jaxbList);
                                }

                            }
                            else {
                                createJaxbList((Collection<?>) methodResult[messageParamIndex],jaxbList);
                            }
                            if (setterMethod != null)
                                setterMethod.invoke(messageJaxbObject,jaxbList);     

                        }

                        else if (messageField.getType().isArray()) {
                            setterMethod.invoke(messageJaxbObject,methodResult[messageParamIndex]);
                        }

                        // The case when the List or the array are wrapped in a JAXB class
                        else {
                            // Get the short class name of the JAXB array class (without package definition)
                            String messagePartShortClassName = messageParts.get(messgePartName).getSimpleName();
                            String createValueMethodName = "create" + messagePartShortClassName;

                            // Get the factory instance to create JAXB array object
                            Class<?> jaxbArrayClass = messageParts.get(messgePartName);
                            Class<?> arrayFactoryClass = jaxbClassLoader.loadClass(
                                    jaxbArrayClass.getPackage().getName() + "." + 
                                    DataBindingConstants.JAXB_FACTORY_CLASS_NAME);
                            Object arrayFactoryInstance = arrayFactoryClass.newInstance();

                            // Get method from the ObjectFactory class to create the JAXB array instance
                            Method[] arrayFactoryMethods = arrayFactoryClass.getDeclaredMethods();
                            Method createArrayMethod = null;
                            for (Method method : arrayFactoryMethods) {
                                if (method.getName().equals(createValueMethodName)) {
                                    createArrayMethod = method;
                                }
                            }
                            jaxbList = createArrayMethod.invoke(arrayFactoryInstance, new Object[]{});
                            if (methodResult[messageParamIndex] != null) {
                                if (!returnType.isArray()) 
                                    createJaxbList((Collection<?>) methodResult[messageParamIndex],jaxbList);
                                else {
                                    try {
                                        createJaxbList(java.util.Arrays.asList(
                                                (Object[]) methodResult[messageParamIndex]),jaxbList);
                                    }
                                    catch(java.lang.ClassCastException e) {
                                        // The array contains a primitive type
                                        List<?> listOfWrappedClass = Util.parimitiveArrayToList(methodResult[messageParamIndex]);
                                        createJaxbList(listOfWrappedClass,jaxbList);
                                    }

                                }
                            }
                            setterMethod.invoke(messageJaxbObject,jaxbList);
                        }

                    }

                    // The method result is a bean
                    else {
                        Object jaxbBean = createJaxblBean(methodResult[messageParamIndex]);
                        Method createMethod = null;
                        if (messageJaxbClass.getDeclaredFields()[messageParamIndex].getType().equals(JAXBElement.class)) 
                            createMethod = jaxbFactoryClass.getDeclaredMethod(createMethodName, 
                                    jaxbBean.getClass());
                        if (createMethod != null)
                            setterMethod.invoke(messageJaxbObject,
                                    createMethod.invoke(jaxbFactoryInstance, jaxbBean));
                        else
                            setterMethod.invoke(messageJaxbObject, jaxbBean);
                    }
                    messageParamIndex ++;
                }
            }
            catch(Exception e) {
                throw new DataBindingException(e);
            }
        }

        try {
            marshaller.setProperty(Marshaller.JAXB_FORMATTED_OUTPUT, Boolean.TRUE);
            if (messageJaxbObject.getClass().getAnnotation(XmlRootElement.class) != null)
                marshaller.marshal(messageJaxbObject, result);

            else if(messageJaxbObject.getClass().getAnnotation(XmlType.class) != null) {
                JAXBElement<?> jaxbElement = new JAXBElement(
                        outputMessage.getMessageName(), 
                        messageJaxbObject.getClass(), 
                        messageJaxbObject );
                marshaller.marshal(jaxbElement, result);
            }

            else 
                throw new DataBindingException(" Can't marshall the " + 
                        outputMessage.getMessageName().getLocalPart() + 
                "Missing Jaxb annotation");
        } catch (Exception e) {
            throw new DataBindingException("Error occured when marshalling the message ",e);
        }



        return result;
    }



    /**
     * Marshall method parameters of the service operation given in parameter
     * @param serviceOperation : The service operation
     * @param methodParams : The method parameters
     * @return : A XML message request to call the operation given in parameter
     * @throws DataBindingException
     */
    @SuppressWarnings("unchecked")
    public Document marshallMethodParam(Operation serviceOperation, Object[] methodParams) throws DataBindingException  {

        DocumentBuilderFactory pFactory = DocumentBuilderFactory.newInstance();
        Document result =null;
        //initializing result
        try {
            result = pFactory.newDocumentBuilder().newDocument();
        } catch (ParserConfigurationException e1) {
            throw new DataBindingException(e1);
        }   

        Message inputMessage = serviceOperation.getInputMessage();
        String messageNameSpace = inputMessage.getMessageTypeName().getNamespaceURI();
        Registry registry = serviceOperation.getServiceOfOperation().getRegistry();
        Marshaller marshaller = registry.getMarshaller(messageNameSpace);
        Object messageJaxbObject = null;

        Class<?> jaxbFactoryClass = registry.getJaxbFactoryInstance(messageNameSpace).getClass();
        Object jaxbFactoryInstance = registry.getJaxbFactoryInstance(messageNameSpace);

        Map<String,Class<?>> messageParts = inputMessage.getMessageParts();
        String methodName = "create" + Util.upperFirstCase(inputMessage.getMessageTypeName().getLocalPart());
        Method createJaxbObject;
        try {
            createJaxbObject = jaxbFactoryClass.getDeclaredMethod(methodName, new Class[]{});
            messageJaxbObject = createJaxbObject.invoke(jaxbFactoryInstance, new Object[]{});
        } catch (Exception e1) {
            throw new DataBindingException("Error occured when creating the message object" , e1);
        } 

        if (serviceOperation.getParameterTypes() != null && 
                serviceOperation.getParameterTypes().length == 1 && 
                serviceOperation.getParameterTypes()[0].getName().
                equalsIgnoreCase(serviceOperation.getInputMessage().getMessageJaxbClass().getName())) {
            Object tempMessageJaxbObject = createJaxblBean(methodParams[0]);
            if (tempMessageJaxbObject != null)
                messageJaxbObject = tempMessageJaxbObject;
        }
        else {

            try {


                Class<?> messageJaxbClass = inputMessage.getMessageJaxbClass();

                Class<?>[] methodParamTypes = serviceOperation.getParameterTypes();

                // Browse the method parameters
                int messageParamIndex = 0;
                for (String messgePartName : messageParts.keySet()) {

                    Class<?> messagePartClass = messageParts.get(messgePartName);
                    Method setterMethod = null;

                    String attributeName = null;
                    if (messageJaxbClass.getDeclaredFields()[messageParamIndex].getAnnotation(XmlElementRef.class) != null
                            && messageJaxbClass.getDeclaredFields()[messageParamIndex].getAnnotation(XmlElementRef.class).name() != null)
                        attributeName = messageJaxbClass.getDeclaredFields()[messageParamIndex].getAnnotation(XmlElementRef.class).name();
                    else 
                        attributeName = messageJaxbClass.getDeclaredFields()[messageParamIndex].getName();
                    String createMethodName = "create" + messageJaxbClass.getSimpleName() 
                    + Util.upperFirstCase(attributeName);


                    // If the field is a collection type, there is no a setter method
                    if (!java.util.Collection.class.isAssignableFrom(messagePartClass)) {
                        String setterMethodName = "set" + Util.upperFirstCase(messgePartName);
                        setterMethod = messageJaxbClass.getDeclaredMethod(setterMethodName, 
                                new Class[]{messagePartClass});
                    }

                    if (methodParams[messageParamIndex] == null) {
                        setterMethod.invoke(messageJaxbObject, (Object) null);
                    }

                    // the current parameter is a simple type
                    else if (methodParamTypes[messageParamIndex].isPrimitive() || Util.isSimpleType(methodParamTypes[messageParamIndex])) {

                        Method createMethod = null;
                        if (messageJaxbClass.getDeclaredFields()[messageParamIndex].getType().equals(JAXBElement.class)) 
                            createMethod = jaxbFactoryClass.getDeclaredMethod(createMethodName, 
                                    methodParamTypes[messageParamIndex]);

                        if (createMethod != null) 
                            setterMethod.invoke(messageJaxbObject, 
                                    createMethod.invoke(jaxbFactoryInstance, methodParams[messageParamIndex]));

                        else
                            setterMethod.invoke(messageJaxbObject, new Object[]{methodParams[messageParamIndex]});
                    }

                    // The current parameter is a Calendar type
                    else if (methodParamTypes[messageParamIndex].getName().equals(DataBindingConstants.CALENDAR_CLASS_NAME)) {

                        javax.xml.datatype.XMLGregorianCalendar jaxbDate = null;
                        java.util.Calendar originalBeanDate = (java.util.Calendar) methodParams[messageParamIndex];
                        if (originalBeanDate != null) {
                            jaxbDate = createXMLGregorianCalendarFromDate(originalBeanDate.getTime());
                        }
                        Method createMethod = null;
                        if (messageJaxbClass.getDeclaredFields()[messageParamIndex].getType().equals(JAXBElement.class)) 
                            createMethod = jaxbFactoryClass.getDeclaredMethod(createMethodName, 
                                    javax.xml.datatype.XMLGregorianCalendar.class);

                        if (createMethod != null) 
                            setterMethod.invoke(messageJaxbObject,
                                    createMethod.invoke(jaxbFactoryInstance,jaxbDate));
                        else
                            setterMethod.invoke(messageJaxbObject, new Object[]{jaxbDate});

                    }

                    // The current parameter is a Date type
                    else if (methodParamTypes[messageParamIndex].getName().equals(DataBindingConstants.DATE_CLASS_NAME) ||
                            methodParamTypes[messageParamIndex].getName().equals(DataBindingConstants.SQL_DATE_CLASS_NAME)) {

                        javax.xml.datatype.XMLGregorianCalendar jaxbDate = null;
                        if (methodParams[messageParamIndex] != null) {
                            jaxbDate = createXMLGregorianCalendarFromDate((java.util.Date) methodParams[messageParamIndex]);
                        }
                        Method createMethod = null;
                        if (messageJaxbClass.getDeclaredFields()[messageParamIndex].getType().equals(JAXBElement.class)) 
                            createMethod = jaxbFactoryClass.getDeclaredMethod(createMethodName, 
                                    javax.xml.datatype.XMLGregorianCalendar.class);

                        if (createMethod != null) 
                            setterMethod.invoke(messageJaxbObject,
                                    createMethod.invoke(jaxbFactoryInstance,jaxbDate));
                        else
                            setterMethod.invoke(messageJaxbObject, new Object[]{jaxbDate});

                    }

                    // The current parameter is a list
                    else if (java.util.Collection.class.isAssignableFrom(methodParamTypes[messageParamIndex]) || 
                            methodParamTypes[messageParamIndex].isArray()) {

                        Object jaxbList = null;
                        Field messageField = Util.getJaxbField(messgePartName, messageJaxbClass.getDeclaredFields());


                        if (java.util.Collection.class.isAssignableFrom(messageField.getType())) {
                            if (setterMethod != null)
                                jaxbList = new ArrayList<Object>();
                            else {
                                String getterMethodName = "get" + Util.upperFirstCase(messgePartName);
                                Method getterMethod = messageJaxbClass.getDeclaredMethod(getterMethodName);
                                jaxbList = getterMethod.invoke(messageJaxbObject);
                            }
                            if (methodParams[messageParamIndex].getClass().isArray()) {
                                try {
                                    createJaxbList(java.util.Arrays.asList((Object[]) methodParams[messageParamIndex]),jaxbList);
                                }
                                catch (java.lang.ClassCastException e) {
                                    // The array contains a primitive type
                                    List<?> listOfWrappedClass = Util.parimitiveArrayToList(methodParams[messageParamIndex]);
                                    createJaxbList(listOfWrappedClass,jaxbList);
                                }

                            }
                            else {
                                createJaxbList((Collection<?>) methodParams[messageParamIndex],jaxbList);
                            }
                            if (setterMethod != null)
                                setterMethod.invoke(messageJaxbObject,jaxbList);    

                        }

                        else if (messageField.getType().isArray()) {
                            setterMethod.invoke(messageJaxbObject,methodParams[messageParamIndex]);
                        }

                        // The case when the List or the array are wrapped in a JAXB class
                        else {

                            // Get the short class name of the JAXB array class (without package definition)
                            String messagePartShortClassName = messageParts.get(messgePartName).getSimpleName();
                            String createValueMethodName = "create" + messagePartShortClassName;

                            // Get the factory instance to create JAXB array object
                            Class<?> jaxbArrayClass = messageParts.get(messgePartName);
                            Class<?> arrayFactoryClass = jaxbClassLoader.loadClass(jaxbArrayClass.getPackage().getName() + "." + 
                                    DataBindingConstants.JAXB_FACTORY_CLASS_NAME);
                            Object arrayFactoryInstance = arrayFactoryClass.newInstance();

                            // Get method from the ObjectFactory class to create the JAXB array instance
                            Method[] arrayFactoryMethods = arrayFactoryClass.getDeclaredMethods();
                            Method createArrayMethod = null;
                            for (Method method : arrayFactoryMethods) {
                                if (method.getName().equals(createValueMethodName)) {
                                    createArrayMethod = method;
                                }
                            }
                            jaxbList = createArrayMethod.invoke(arrayFactoryInstance, new Object[]{});
                            if (methodParams[messageParamIndex] != null) {
                                if (!methodParamTypes[messageParamIndex].isArray())
                                    createJaxbList((Collection<?>) methodParams[messageParamIndex], jaxbList);
                                else {
                                    try {
                                        createJaxbList(java.util.Arrays.asList(
                                                (Object[]) methodParams[messageParamIndex]),jaxbList);
                                    }
                                    catch(java.lang.ClassCastException e) {
                                        // The array contains a primitive type
                                        List<?> listOfWrappedClass = Util.parimitiveArrayToList(methodParams[messageParamIndex]);
                                        createJaxbList(listOfWrappedClass,jaxbList);
                                    }

                                }
                            }
                            setterMethod.invoke(messageJaxbObject,jaxbList);
                        }

                    }
                    // The current parameter is a bean
                    else {
                        Object jaxbBean = createJaxblBean(methodParams[messageParamIndex]);
                        Method createMethod = null;
                        if (messageJaxbClass.getDeclaredFields()[messageParamIndex].getType().equals(JAXBElement.class)) 
                            createMethod = jaxbFactoryClass.getDeclaredMethod(createMethodName, 
                                    jaxbBean.getClass());
                        if (createMethod != null)
                            setterMethod.invoke(messageJaxbObject,
                                    createMethod.invoke(jaxbFactoryInstance, jaxbBean));
                        else
                            setterMethod.invoke(messageJaxbObject, jaxbBean);

                    }
                    messageParamIndex ++;
                }
            }
            catch (Exception e) {
                throw new DataBindingException(e);
            }
        }

        try {
            marshaller.setProperty(Marshaller.JAXB_FORMATTED_OUTPUT, Boolean.TRUE);
            if (messageJaxbObject.getClass().getAnnotation(XmlRootElement.class) != null)
                marshaller.marshal(messageJaxbObject, result);

            else if(messageJaxbObject.getClass().getAnnotation(XmlType.class) != null) {
                JAXBElement<?> jaxbElement = new JAXBElement(
                        inputMessage.getMessageName(), 
                        messageJaxbObject.getClass(), 
                        messageJaxbObject );
                marshaller.marshal(jaxbElement, result);
            }
        } catch (Exception e) {
            throw new DataBindingException("Error occured when marshalling the message ",e);
        }

        return result;
    }


    /**
     * 
     * @param serviceOperation
     * @param exception
     * @throws DataBindingException
     */
    @SuppressWarnings("unchecked")
    public Document marshallMethodException(Operation serviceOperation, Exception exception) throws DataBindingException  {
        
        //initializing result
        DocumentBuilderFactory pFactory = DocumentBuilderFactory.newInstance();
        Document result =null;
        try {
            result = pFactory.newDocumentBuilder().newDocument();
        } catch (ParserConfigurationException e1) {
            throw new DataBindingException(e1);
        }   

        // Fetch the WSDL fault matching with the exception
        Message faultMessage = null;
        List<Message> faults = serviceOperation.getOperationFaults();
        for(Message fault : faults) {
            if (fault.getMessageName().getLocalPart().equalsIgnoreCase(exception.getClass().getSimpleName()) ||
                    fault.getMessageTypeName().getLocalPart().equalsIgnoreCase(exception.getClass().getSimpleName()))
                faultMessage = fault;
        }


        if (faultMessage == null)
            throw new DataBindingException(" The Exception : " + exception.getClass().getName() + 
            " is not defined in the WSDL ");


        String messageNameSpace = faultMessage.getMessageTypeName().getNamespaceURI();
        Registry registry = serviceOperation.getServiceOfOperation().getRegistry();
        Marshaller marshaller = registry.getMarshaller(messageNameSpace);
        

        Class<?> exceptionClass = exception.getClass();
        Field[] fields = exceptionClass.getDeclaredFields();
        Object jaxbException = null;
        if (fields != null && fields.length == 1 &&
                fields[0].getType().getSimpleName().equalsIgnoreCase(
                        faultMessage.getMessageTypeName().getLocalPart())) {

            try {
                jaxbException = createJaxblBean(exceptionClass.getDeclaredMethod(
                        "get" + Util.upperFirstCase(fields[0].getName())).invoke(exception));
            } catch (Exception e) {
               throw new DataBindingException(" Error occured while creating the JAXB Exception ",e);
            } 

        }
        else {
            jaxbException = createJaxblBean(exception);
        }
        
        JAXBElement<?> jaxbElement;
        if (jaxbException != null) {
            jaxbElement = new JAXBElement(
                    faultMessage.getMessageName(), 
                    jaxbException.getClass(), 
                    jaxbException );
        }
        else  {
            Class<?> jaxbFactoryClass = registry.getJaxbFactoryInstance(messageNameSpace).getClass();
            Object jaxbFactoryInstance = registry.getJaxbFactoryInstance(messageNameSpace);

            String methodName = "create" + Util.upperFirstCase(faultMessage.getMessageTypeName().getLocalPart());
            try {
                Method createJaxbObject = jaxbFactoryClass.getDeclaredMethod(methodName, new Class[]{});
                Object messageJaxbObject = createJaxbObject.invoke(jaxbFactoryInstance, new Object[]{});
                jaxbElement = new JAXBElement(faultMessage.getMessageName(), messageJaxbObject.getClass(), 
                        messageJaxbObject );
            } catch (Exception e1) {
                throw new DataBindingException("Error occured when creating the message object" , e1);
            } 
        }
        
        try {
            marshaller.marshal(jaxbElement, result);
        } catch (Exception e) {
            throw new DataBindingException(" Error occured when marshalling the databinding classe ",e);
        } 
        

       return result;

    }




    /**
     * 
     * @param originalBeanInstance
     * @param originalBeanParents
     * @param jaxbBeanParents
     * @return
     * @throws DataBindingException
     */
    private Object createJaxblBean(Object originalBeanInstance)  throws DataBindingException {

        if (originalBeanInstance == null) return null;

        Object jaxbBeanInstance = null;
        try {

            String beanClassName = originalBeanInstance.getClass().getSimpleName();
            // Load the JAXB class for the current bean
            String beanPackageName = getClassPackage(beanClassName);
            if ("".equals(beanPackageName)){
                beanPackageName = getClassPackage(Util.lowerFirstCase(beanClassName));
            }
            if ("".equals(beanPackageName)){
                beanPackageName = originalBeanInstance.getClass().getPackage().getName();
            }

            // Get the correct class name in the case when the bean is an inner class
            beanClassName = originalBeanInstance.getClass().getName().replace(
                    originalBeanInstance.getClass().getPackage().getName()+".", "");

            Class<?> jaxbBeanClass = null;
            String jaxbBeanClassName = null;
            try {
                jaxbBeanClassName = beanPackageName + "." + beanClassName;
                jaxbBeanClass = jaxbClassLoader.loadClass(jaxbBeanClassName);
            }
            catch(java.lang.NoClassDefFoundError e) {
                // Can be that the package name contains uppercase characters
                beanPackageName = beanPackageName.toLowerCase();
                jaxbBeanClassName = beanPackageName + "." + beanClassName;
                jaxbBeanClass = jaxbClassLoader.loadClass(jaxbBeanClassName);
            }


            String simpleClassName = jaxbBeanClass.getCanonicalName().replace(
                    jaxbBeanClass.getPackage().getName(), "");
            // TO DO : Add regular expression to replace all '.' character.
            simpleClassName = simpleClassName.replace(".", "");
            String methodName = "create" + simpleClassName;

            Class<?> beanFactoryClass = jaxbClassLoader.loadClass(beanPackageName 
                    + "." + DataBindingConstants.JAXB_FACTORY_CLASS_NAME);
            Object beanFactoryInstance = beanFactoryClass.newInstance();
            Method createJaxbObject = beanFactoryClass.getDeclaredMethod(methodName, new Class[]{});
            jaxbBeanInstance = createJaxbObject.invoke(beanFactoryInstance, new Object[]{});

            Field[] originalBeanFields = Util.getFieldFormBeanClassAndSuperClass(originalBeanInstance.getClass());
            for (Field originalBeanField : originalBeanFields) {

                // Initialize the field factory class and instance
                Class<?> fieldFactoryClass = beanFactoryClass;
                Object fieldFactoryInstance = beanFactoryInstance; 

                if (originalBeanField.getName().equals("serialVersionUID")) 
                    continue; 

                Field[] jaxbBeanFields = Util.getFieldFormBeanClassAndSuperClass(jaxbBeanClass);
                Field jaxbBeanField  = Util.getJaxbField(originalBeanField.getName(), jaxbBeanFields);

                if (jaxbBeanField == null)
                    continue;

                Class<?> declaringFieldClass = jaxbBeanField.getDeclaringClass();
                // Check if the current field is an attribute of the super class
                if (!declaringFieldClass.getPackage().getName().equals(beanFactoryClass.getPackage().getName())) {
                    fieldFactoryClass = jaxbClassLoader.loadClass(declaringFieldClass.getPackage().getName() 
                            + "." + DataBindingConstants.JAXB_FACTORY_CLASS_NAME);
                    fieldFactoryInstance = fieldFactoryClass.newInstance();
                }

                String getterMethodName;
                if (!originalBeanField.getType().equals(boolean.class) 
                        && !originalBeanField.getType().equals(java.lang.Boolean.class)) {
                    getterMethodName = "get" + Util.upperFirstCase(originalBeanField.getName());
                }
                else {
                    getterMethodName = "is" + Util.upperFirstCase(originalBeanField.getName());
                }


                Method getterMethod = null;
                getterMethod = Util.getMethodFomBeanClassOrSuperClass(getterMethodName, 
                        originalBeanInstance.getClass());

                if (getterMethod == null && getterMethodName.startsWith("is")) {
                    getterMethodName = getterMethodName.replaceFirst("is", "get");
                    getterMethod = Util.getMethodFomBeanClassOrSuperClass(getterMethodName, 
                            originalBeanInstance.getClass());
                }

                if (getterMethod == null) 
                    throw new DataBindingException(" No such method  " + getterMethod);
                

                String setterMethodName = null;
                Method setterMethod  = null;
                XmlElement xmlElement = jaxbBeanField.getAnnotation(XmlElement.class);
                // The setter method name is deducted from the xmlElement 
                // annotation if exists 
                if (xmlElement != null && xmlElement.name() != null) {
                    setterMethodName = "set" + Util.upperFirstCase(xmlElement.name());
                }
                else {
                    setterMethodName = "set" + Util.upperFirstCase(jaxbBeanField.getName());
                }
                
                try {
                    // If the field is a collection type, there is no a setter method
                    if (!java.util.Collection.class.isAssignableFrom(jaxbBeanField.getType()))
                        setterMethod = jaxbBeanClass.getMethod(setterMethodName, 
                                new Class[]{jaxbBeanField.getType()});
                }
                catch(java.lang.NoSuchMethodException e) {
                    setterMethodName = "set" + Util.upperFirstCase(originalBeanField.getName());
                    setterMethod = jaxbBeanClass.getMethod(setterMethodName, 
                            new Class[]{jaxbBeanField.getType()});
                }

                String createValueMethodName = "create" + declaringFieldClass.getSimpleName() 
                + Util.upperFirstCase(jaxbBeanField.getName());

                // No treatment when the attribute value is null to avoid the serialization of 
                // the null elements in the XML message
                if (getterMethod.invoke(originalBeanInstance, new Object[]{}) == null)
                    continue;

                // The Current field is a primitive type
                if (originalBeanField.getType().isPrimitive() || Util.isSimpleType(originalBeanField.getType())) {

                    if (jaxbBeanField.getType().equals(javax.xml.bind.JAXBElement.class)) {
                        Method createValueMethod = fieldFactoryClass.getDeclaredMethod(createValueMethodName, 
                                new Class[]{originalBeanField.getType()});
                        Object jaxbFieldValue = createValueMethod.invoke(fieldFactoryInstance, 
                                getterMethod.invoke(originalBeanInstance, new Object[]{}));
                        setterMethod.invoke(jaxbBeanInstance, new Object[]{jaxbFieldValue});
                    }
                    else 
                        setterMethod.invoke(jaxbBeanInstance, 
                                getterMethod.invoke(originalBeanInstance, new Object[]{}));

                }


                // The current parameter is a Calendar type
                else if (originalBeanField.getType().getName().
                        equals(DataBindingConstants.CALENDAR_CLASS_NAME)) {

                    javax.xml.datatype.XMLGregorianCalendar jaxbDate = null;
                    java.util.Calendar originalBeanDate = 
                        (java.util.Calendar) getterMethod.invoke(originalBeanInstance, new Object[]{});
                    Object jaxbFieldValue = null;
                    if (originalBeanDate != null) {
                        jaxbDate = createXMLGregorianCalendarFromDate(originalBeanDate.getTime());
                        Method createValueMethod = fieldFactoryClass.getDeclaredMethod(createValueMethodName, 
                                new Class[]{javax.xml.datatype.XMLGregorianCalendar.class});
                        jaxbFieldValue = createValueMethod.invoke(fieldFactoryInstance, jaxbDate);

                    }
                    setterMethod.invoke(jaxbBeanInstance, new Object[]{jaxbFieldValue});

                }

                // The current parameter is a Date type
                else if (originalBeanField.getType().getName().equals(DataBindingConstants.DATE_CLASS_NAME) ||
                        originalBeanField.getType().getName().equals(DataBindingConstants.SQL_DATE_CLASS_NAME)) {

                    javax.xml.datatype.XMLGregorianCalendar jaxbDate = null;
                    java.util.Date originalBeanDate = 
                        (java.util.Date) getterMethod.invoke(originalBeanInstance, new Object[]{});
                    if (originalBeanDate != null) {
                        jaxbDate = createXMLGregorianCalendarFromDate(originalBeanDate);
                    }
                    if (jaxbBeanField.getType().equals(javax.xml.bind.JAXBElement.class)) {
                        Method createValueMethod = fieldFactoryClass.getDeclaredMethod(createValueMethodName, 
                                new Class[]{javax.xml.datatype.XMLGregorianCalendar.class});
                        setterMethod.invoke(jaxbBeanInstance, createValueMethod.invoke(fieldFactoryInstance, jaxbDate));

                    }
                    else 
                        setterMethod.invoke(jaxbBeanInstance, new Object[]{jaxbDate});

                }

                // The Current field is a List
                else if (java.util.Collection.class.isAssignableFrom(originalBeanField.getType()) 
                        || originalBeanField.getType().isArray()) {

                    Object originalBeanList = getterMethod.invoke(originalBeanInstance, new Object[]{});
                    // If the field is null, nothing to do
                    if (originalBeanList == null)
                        continue;

                    if (java.util.Collection.class.isAssignableFrom(jaxbBeanField.getType())) {
                        List<?> jaxbList = null;
                        if (setterMethod != null)
                            jaxbList = new ArrayList<Object>();
                        else {
                            String jaxbGetterMethodName = "get" + Util.upperFirstCase(jaxbBeanField.getName());
                            Method jaxbgetterMethod = jaxbBeanClass.getMethod(jaxbGetterMethodName);
                            jaxbList = (List<?>) jaxbgetterMethod.invoke(jaxbBeanInstance);
                        }
                        if (originalBeanList.getClass().isArray()) {
                            try {
                                createJaxbList(java.util.Arrays.asList((Object[]) originalBeanList),jaxbList);
                            }
                            catch (java.lang.ClassCastException e) {
                                // The array contains a primitive type
                                List<?> listOfWrappedClass = Util.parimitiveArrayToList(originalBeanList);
                                createJaxbList(listOfWrappedClass,jaxbList);
                            }

                        }
                        else {
                            createJaxbList((Collection<?>) originalBeanList,jaxbList);
                        }
                        if (setterMethod != null)
                            setterMethod.invoke(jaxbBeanInstance,jaxbList);    
                    }

                    else if (jaxbBeanField.getType().isArray()) {
                        setterMethod.invoke(jaxbBeanInstance,originalBeanList);
                    }

                    // The case when the List or the array are wrapped in a JAXB class
                    else {

                        Method[] jaxbBeanMethods = fieldFactoryClass.getDeclaredMethods();
                        Method createValueMethod = null;
                        for (Method method : jaxbBeanMethods) {
                            if (method.getName().equals(createValueMethodName)) {
                                createValueMethod = method;
                            }
                        }


                        Class<?> jaxbArray = createValueMethod.getParameterTypes()[0];

                        if (jaxbArray.isArray()) {
                            Object jaxbFieldValue = createValueMethod.invoke(fieldFactoryInstance, 
                                    getterMethod.invoke(originalBeanInstance, new Object[]{}));
                            setterMethod.invoke(jaxbBeanInstance, jaxbFieldValue);

                        }
                        else {
                            String jaxbArrayFactoryClassName = jaxbArray.getPackage().getName() + "." + DataBindingConstants.JAXB_FACTORY_CLASS_NAME;
                            Class<?> jaxbArrayFactoryClass = jaxbClassLoader.loadClass(jaxbArrayFactoryClassName);
                            Object jaxbArrayFactoryInstance = jaxbArrayFactoryClass.newInstance();

                            String jaxbArrayClassName = jaxbArray.getSimpleName();
                            String createJaxbArrayMethodName = "create" + jaxbArrayClassName;
                            Method createJaxbArrayMethod = jaxbArrayFactoryClass.getDeclaredMethod(createJaxbArrayMethodName,new Class[]{});
                            Object jaxbArrayObject = createJaxbArrayMethod.invoke(jaxbArrayFactoryInstance, new Object[]{});

                            if (!originalBeanField.getType().isArray())
                                createJaxbList((Collection<?>)  originalBeanList,jaxbArrayObject);
                            else {
                                createJaxbList(java.util.Arrays.asList((Object[]) originalBeanList),jaxbArrayObject);
                            }

                            Object jaxbFieldValue = createValueMethod.invoke(fieldFactoryInstance, jaxbArrayObject);
                            setterMethod.invoke(jaxbBeanInstance, jaxbFieldValue);
                        }
                    }
                }

                else if (originalBeanField.getType().isEnum()) {

                    if (jaxbBeanField.getType().equals(javax.xml.bind.JAXBElement.class)) {
                        Method createValueMethod = fieldFactoryClass.getDeclaredMethod(createValueMethodName, 
                                new Class[]{originalBeanField.getType()});
                        Object jaxbFieldValue = createValueMethod.invoke(fieldFactoryInstance, 
                                createEnumObject(getterMethod.invoke(originalBeanInstance, new Object[]{}), jaxbBeanField.getType()));
                        setterMethod.invoke(jaxbBeanInstance, new Object[]{jaxbFieldValue});
                    }
                    else 
                        setterMethod.invoke(jaxbBeanInstance, 
                                createEnumObject(getterMethod.invoke(originalBeanInstance, new Object[]{}), jaxbBeanField.getType()));

                }

                else {



                    Object jaxbBean = 
                        createJaxblBean(getterMethod.invoke(originalBeanInstance, new Object[]{}));
                    if (jaxbBean != null) {
                        if (!jaxbBeanField.getType().equals(javax.xml.bind.JAXBElement.class)) {
                            setterMethod.invoke(jaxbBeanInstance, jaxbBean);
                        }
                        // The JAXB field is wrapped in JAXBElement class
                        else {
                            Method createValueMethod = 
                                fieldFactoryClass.getDeclaredMethod(createValueMethodName, jaxbBean.getClass());
                            Object jaxbFieldValue = 
                                createValueMethod.invoke(fieldFactoryInstance, jaxbBean);
                            setterMethod.invoke(jaxbBeanInstance, new Object[]{jaxbFieldValue});
                        }
                    }
                    else 
                        setterMethod.invoke(jaxbBeanInstance, new Object[]{null});

                }

            }
        } catch (Exception e) {
            throw new DataBindingException(e);
        } 

        return jaxbBeanInstance;
    }

    /**
     * 
     * @param jaxbEnumInstance
     * @param originalEnumClazz
     * @return
     * @throws Exception
     */
    private Object createEnumObject(Object originalEnumInstance, Class<?> jaxbEnumClazz) throws Exception {
        Class<?> jaxbEnumClass = null;
        try {
            jaxbEnumClass = jaxbClassLoader.loadClass(originalEnumInstance.getClass().getName());
        }
        catch (Exception e) {
            // The case when the name space and the package name are different
            if (jaxbEnumClazz != null)
                jaxbEnumClass = jaxbClassLoader.loadClass(jaxbEnumClazz.getName());
            else 
                throw new DataBindingException(e);
        }

        Method valueMethod = originalEnumInstance.getClass().getDeclaredMethod("value"); 
        Method fromValueMethod = jaxbEnumClass.getDeclaredMethod("fromValue",valueMethod.invoke(originalEnumInstance).getClass()) ;
        return fromValueMethod.invoke(jaxbEnumClass,valueMethod.invoke(originalEnumInstance));
    }







    /**
     * 
     * @param originalList
     * @param jaxbArryInsyance
     * @param originalBeanParents
     * @param jaxbBeanParents
     * @throws DataBindingException
     */
    @SuppressWarnings({ "unchecked"})
    private void createJaxbList(Collection<?> originalList, Object jaxbArryInsyance) throws DataBindingException {

        try {

            Collection<Object> jaxbList ;
            if (!(jaxbArryInsyance instanceof Collection)) {
                jaxbList = 
                    (Collection<Object>) jaxbArryInsyance.getClass().getDeclaredMethods()[0].invoke(jaxbArryInsyance, new Object[0]);
            }
            else
                jaxbList = (Collection<Object>) jaxbArryInsyance;


            for (Object listElement : originalList) {

                if (listElement == null || Util.isSimpleType(listElement.getClass())) {
                    jaxbList.add(listElement);
                }


                // The current element list is a Calendar type
                else if (listElement.getClass().getName().equals(DataBindingConstants.CALENDAR_CLASS_NAME)) {

                    javax.xml.datatype.XMLGregorianCalendar jaxbDate = null;
                    java.util.Calendar originalDate = (java.util.Calendar) listElement;
                    if (originalDate != null) {
                        jaxbDate = createXMLGregorianCalendarFromDate(originalDate.getTime());

                    }
                    jaxbList.add(jaxbDate);

                }

                // The current element list is a Date type
                else if (listElement.getClass().getName().equals(DataBindingConstants.DATE_CLASS_NAME)) {

                    javax.xml.datatype.XMLGregorianCalendar jaxbDate = null;
                    java.util.Date originalDate = (java.util.Date) listElement;
                    if (originalDate != null) {
                        jaxbDate = createXMLGregorianCalendarFromDate(originalDate);

                    }
                    jaxbList.add(jaxbDate);

                }

                // The current element list is a collection type
                else if(java.util.Collection.class.isAssignableFrom(listElement.getClass())) {

                    Field field = jaxbArryInsyance.getClass().getDeclaredFields()[0];
                    Type type = field.getGenericType();

                    if (type instanceof ParameterizedType) {

                        ParameterizedType pt = (ParameterizedType) type;
                        Type genericType = pt.getActualTypeArguments()[0];

                        Class<?> clazz = null;

                        if (genericType instanceof Class<?>) 
                            clazz = (Class<?>) genericType;

                        else if (genericType instanceof ParameterizedType) 
                            clazz = (Class<?>) ((ParameterizedType) genericType).getRawType();

                        if (clazz != null) {
                            String jaxbArrayFactoryClassName = clazz.getPackage().getName() + "." + DataBindingConstants.JAXB_FACTORY_CLASS_NAME;
                            Class<?> jaxbArrayFactoryClass = jaxbClassLoader.loadClass(jaxbArrayFactoryClassName);
                            Object jaxbArrayFactoryInstance = jaxbArrayFactoryClass.newInstance();

                            String jaxbArrayClassName = clazz.getSimpleName();
                            String createJaxbArrayMethodName = "create" + jaxbArrayClassName;
                            Method createJaxbArrayMethod = jaxbArrayFactoryClass.getDeclaredMethod(createJaxbArrayMethodName,new Class[]{});
                            Object jaxbArrayObject = createJaxbArrayMethod.invoke(jaxbArrayFactoryInstance, new Object[]{});

                            Collection<?> originalList_ = (Collection<?>) listElement;
                            if (originalList_ != null) {

                                createJaxbList(originalList_,jaxbArrayObject);
                            }
                            jaxbList.add(jaxbArrayObject);  


                        }
                        else throw new DataBindingException(" The element list type : " + 
                                listElement.getClass().getName() + "is not supported ");
                    }

                }

                // The current element list is a bean
                else {

                    Object jaxbBeanInstance = createJaxblBean(listElement);
                    jaxbList.add(jaxbBeanInstance);
                }
            }
        }
        catch(Exception e) {
            throw new DataBindingException(e);
        }
    }

    /**
     * 
     * @param date
     * @return
     * @throws DatatypeConfigurationException
     */
    private XMLGregorianCalendar createXMLGregorianCalendarFromDate(Date date) throws DatatypeConfigurationException {

        GregorianCalendar gCalendar = new GregorianCalendar();
        gCalendar.setTime(date);
        XMLGregorianCalendar jaxbDate = DatatypeFactory.newInstance().newXMLGregorianCalendar(gCalendar);
        return jaxbDate;

    }



    public ClassLoader getJaxbClassLoader() {
        return jaxbClassLoader;
    }



    public void setJaxbClassLoader(ClassLoader jaxbClassLoader) {
        this.jaxbClassLoader = jaxbClassLoader;
    }


    public void setClassPackageMapping(Map<String, String> classPackageMapping) {
        this.classPackageMapping = classPackageMapping;
    }



    public String getClassPackage(String className){
        if (this.classPackageMapping != null && this.classPackageMapping.containsKey(className))
        {
            return this.classPackageMapping.get(className);
        }

        return "";        
    }

}
